#ifndef __AVLTREE_H__
#define __AVLTREE_H__

#include <cstdlib>
#include <vector>
#include <stack>
#include <functional>
#include <cassert>

#define REPLACED_BY_PREDECESSOR_IN_ERASE    1

/**
 * @brief Node type of a balanced binary search tree(AVL tree).
 * 
 * @tparam _Key  Type of key objects.
 * @tparam _Value  Type of mapped objects.
 */
template <typename _Key, typename _Value>
class AVLNode
{
public:
	AVLNode() : key(), value(), left(NULL), right(NULL), BF(0) {}
	AVLNode(_Key k, _Value v) : key(k), value(v), left(NULL), right(NULL), BF(0) {}

	_Key key;     ///< key of this node.
	_Value value; ///< value of this node.
	
	AVLNode *left;   ///< pointing to the left child node of this node.
	AVLNode *right;  ///< pointing to the right child node of this node.

	int BF; ///< balance factor
};

/**
 * @brief A balanced binary search tree(AVL tree), which can be retrieved based on a key, in logarithmic time.
 * 
 * @tparam _Key  Type of key objects.
 * @tparam _Value  Type of mapped objects.
 * @tparam _Compare  Comparison function object type, defaults to less<_Key>.
 */
template <typename _Key, typename _Value, typename _Compare = std::less<_Key> >
class AVLTree
{
public:
	/** @name constructors, destructor of AVLTree. */
	///@{
	AVLTree() : node_count(0), key_compare(), root(NULL) {}
	AVLTree(const _Compare& _compare) : node_count(0), key_compare(), root(NULL) {}
	AVLTree(const AVLTree& rhs) : node_count(rhs.node_count), key_compare(rhs.key_compare), root(rhs.root) {}
	~AVLTree() { clear(); }
	///@}
	
	AVLTree& operator=(const AVLTree& rhs)
	{
		node_count = rhs.node_count;
		key_compare = rhs.key_compare;
		root = rhs.root;
		return *this;
	}

	/** @name APIs of AVLTree. */
	///@{
	/** @brief return true if AVLTree has no nodes. */
	bool empty() { return node_count == 0; }
	/** @brief return the number of nodes of AVLTree. */
	size_t size() { return node_count; }
	/** @brief return the left-most node of AVLTree. */
	AVLNode<_Key, _Value>* minimum();
	/** @brief return the right-most node of AVLTree. */
	AVLNode<_Key, _Value>* maximum();
	/** @brief return the point to the finding node if found, return NULL if not found. */
	AVLNode<_Key, _Value>* find(_Key k);
	/**
	 * @brief insert a node.
	 * 
	 * @param k key of AVLNode.
	 * @param v value of AVLNode.
	 * @return point to the inserted AVLode.
	 */
	AVLNode<_Key, _Value>* insert(_Key k, _Value v);
	/** @brief erase a node only need adjust points. */
	void erase(_Key k);
	/** @brief erase all the nodes of AVLTree. */
	void clear();
	///@}
	
	/** @name Traverse methods of AVLTree. */
	///@{
	void preOrderTraverse(std::vector<AVLNode<_Key, _Value>* >& vec);
	void inOrderTraverse(std::vector<AVLNode<_Key, _Value>* >& vec);
	void postOrderTraverse(std::vector<AVLNode<_Key, _Value>* >& vec);
	///@}

private:
	/** @name Critical methods of AVLTree. */
	///@{
	/** @brief insert a node(recursively). */
	bool _insert(AVLNode<_Key, _Value> **p, AVLNode<_Key, _Value> *node, bool& taller);
	/** @brief erase a node(recursively). */
	bool _erase(AVLNode<_Key, _Value> **p, _Key& k, bool& taller);
	/** @brief erase a node which has both left and right children(recursively). */
	void _erase(AVLNode<_Key, _Value> *q, AVLNode<_Key, _Value> **r, bool& taller);
	/** @brief rebalance left subtree of the erased node. */
	void leftRebalance(AVLNode<_Key, _Value> **p, bool& taller);
	/** @brief rebalance right subtree of the erased node. */
	void rightRebalance(AVLNode<_Key, _Value> **p, bool& taller);
	/** @brief execute LL rotate as to node p. */
	AVLNode<_Key, _Value>* rotateLL(AVLNode<_Key, _Value> **p);
	/** @brief execute RR rotate as to node p. */
	AVLNode<_Key, _Value>* rotateRR(AVLNode<_Key, _Value> **p);
	/** @brief execute LR rotate as to node p. */
	AVLNode<_Key, _Value>* rotateLR(AVLNode<_Key, _Value> **p);
	/** @brief execute RL rotate as to node p. */
	AVLNode<_Key, _Value>* rotateRL(AVLNode<_Key, _Value> **p);
	///@}
	
private:
	size_t node_count;           ///< maintaining the number of nodes of AVLTree.
	_Compare key_compare;        ///< key compare predicate of AVLNode.
	AVLNode<_Key, _Value> *root; ///< root node of AVLTree.
};

//////////////////////////////////////////////////////////////////////////
////////////////////          Implementation          ////////////////////
//////////////////////////////////////////////////////////////////////////
template <typename _Key, typename _Value, typename _Compare>
AVLNode<_Key, _Value>* AVLTree<_Key, _Value, _Compare>::minimum()
{
	if (root == NULL) return NULL;

	AVLNode<_Key, _Value> *p = root;
	while (p->left != NULL)
		p = p->left;
	return p;
}

template <typename _Key, typename _Value, typename _Compare>
AVLNode<_Key, _Value>* AVLTree<_Key, _Value, _Compare>::maximum()
{
	if (root == NULL) return NULL;

	AVLNode<_Key, _Value> *p = root;
	while (p->right != NULL)
		p = p->right;
	return p;
}

template <typename _Key, typename _Value, typename _Compare>
AVLNode<_Key, _Value>* AVLTree<_Key, _Value, _Compare>::find(_Key k)
{
	if (root == NULL) return NULL;

	AVLNode<_Key, _Value> *p = root;
	while (p != NULL)
	{
		if (key_compare(k, p->key)) // k < p->key
			p = p->left;
		else if (key_compare(p->key, k)) // k > p->key
			p = p->right;
		else // k == p->key
			break;
	}
	return p;
}

template <typename _Key, typename _Value, typename _Compare>
AVLNode<_Key, _Value>* AVLTree<_Key, _Value, _Compare>::insert(_Key k, _Value v)
{
	AVLNode<_Key, _Value>* node = new AVLNode<_Key, _Value>(k, v);

	bool taller = false;
	bool hasInsert = _insert(&root, node, taller);
	if (hasInsert)
		++node_count;

	return hasInsert ? node : NULL;
}

template <typename _Key, typename _Value, typename _Compare>
void AVLTree<_Key, _Value, _Compare>::erase(_Key k)
{
	bool taller = false;
	bool hasErase = _erase(&root, k, taller);

	if (hasErase)
		--node_count;
}

template <typename _Key, typename _Value, typename _Compare>
void AVLTree<_Key, _Value, _Compare>::clear()
{
	if (node_count == 0) return;

	std::stack<AVLNode<_Key, _Value>* > S;
	AVLNode<_Key, _Value> *p = root, *q = NULL;
	while (p != NULL || !S.empty())
	{
		if (p != NULL)
		{
			S.push(p);
			p = p->left;
		}
		else
		{
			p = S.top();
			S.pop();
			q = p;
			p = p->right;
			delete q;
		}
	}
	
	node_count = 0;
	root = NULL;
}

template <typename _Key, typename _Value, typename _Compare>
bool AVLTree<_Key, _Value, _Compare>::_insert(AVLNode<_Key, _Value> **p, AVLNode<_Key, _Value> *node, bool& taller)
{
	if ((*p) == NULL)
	{
		(*p) = node;
		taller = true;
		return true;
	}

	if (key_compare(node->key, (*p)->key)) // node->key < p->key
	{
		if (_insert(&((*p)->left), node, taller) == false)
			return false;
		if (taller)
		{
			switch ((*p)->BF)
			{
			case 1:
			{
				(*p)->BF = 2;
				if ((*p)->left->BF == 1)
					*p = rotateLL(p);
				else // p->left->BF == -1
					*p = rotateLR(p);
				taller = false;
				break;
			}
			case 0:
			{
				(*p)->BF = 1;
				taller = true;
				break;
			}
			case -1:
			{
				(*p)->BF = 0;
				taller = false;
				break;
			}
			}
		}
	}
	else if (key_compare((*p)->key, node->key)) // node->key > p->key
	{
		if (_insert(&((*p)->right), node, taller) == false)
			return false;
		if (taller)
		{
			switch ((*p)->BF)
			{
			case 1:
			{
				(*p)->BF = 0;
				taller = false;
				break;
			}
			case 0:
			{
				(*p)->BF = -1;
				taller = true;
				break;
			}
			case -1:
			{
				(*p)->BF = -2;
				if ((*p)->right->BF == -1)
					*p = rotateRR(p);
				else // p->right->BF == 1
					*p = rotateRL(p);
				taller = false;
				break;
			}
			}
		}
	}
	else // node->key == p->key
	{
		delete node;
		taller = false;
		return false;
	}
	return true;
}

template <typename _Key, typename _Value, typename _Compare>
bool AVLTree<_Key, _Value, _Compare>::_erase(AVLNode<_Key, _Value> **p, _Key& k, bool& taller)
{
	if (*p == NULL)
		return false;

	if (key_compare(k, (*p)->key)) // k < p->key
	{
		if (_erase(&((*p)->left), k, taller) == false)
			return false;
		if (taller)
			leftRebalance(p, taller);
	}
	else if (key_compare((*p)->key, k)) // k > p->key
	{
		if (_erase(&((*p)->right), k, taller) == false)
			return false;
		if (taller)
			rightRebalance(p, taller);
	}
	else // k == p->key
	{
		AVLNode<_Key, _Value> *q = *p;
		if ((*p)->left == NULL)
		{
			(*p) = (*p)->right;
			delete q;
			taller = true;
		}
		else if ((*p)->right == NULL)
		{
			(*p) = (*p)->left;
			delete q;
			taller = true;
		}
		else // (*p)->left != NULL && (*p)->right != NULL
		{
#if REPLACED_BY_PREDECESSOR_IN_ERASE
			_erase(q, &q->left, taller);
			if (taller)
				leftRebalance(&q, taller);
#else
			_erase(q, &q->right, taller);
			if (taller)
				rightRebalance(&q, taller);
#endif
			*p = q;
		}
	}
	return true;
}

template <typename _Key, typename _Value, typename _Compare>
void AVLTree<_Key, _Value, _Compare>::_erase(AVLNode<_Key, _Value> *q, AVLNode<_Key, _Value> **r, bool& taller)
{
#if REPLACED_BY_PREDECESSOR_IN_ERASE
	if ((*r)->right == NULL) // r is the predecessor of the erased node q
	{
		q->key = (*r)->key;
		q->value = (*r)->value;
		q = *r; // pointer q is passed by value, needn't worry about modifying it
		*r = (*r)->left; // make (*r)->parent->right = (*r)->left
		delete q;
		taller = true; // after erasing the predecessor of node q, height decreased by 1
	}
	else
	{
		_erase(q, &(*r)->right, taller); // equivalent to always execute *r = (*r)->right, indicating to find the right-most node in the left-subtree of q
		if (taller)
			rightRebalance(r, taller);
	}
#else
	if ((*r)->left == NULL) // r is the successor of the erased node q
	{
		q->key = (*r)->key;
		q->value = (*r)->value;
		q = *r; // pointer q is passed by value, needn't worry about modifying it
		*r = (*r)->right; // make (*r)->parent->left = (*r)->right
		delete q;
		taller = true; // after erasing the successor of node q, height decreased by 1
	}
	else
	{
		_erase(q, &(*r)->left, taller); // equivalent to always execute *r = (*r)->left, indicating to find the left-most node in the right-subtree of q
		if (taller)
			leftRebalance(r, taller);
	}
#endif
}

template <typename _Key, typename _Value, typename _Compare>
void AVLTree<_Key, _Value, _Compare>::leftRebalance(AVLNode<_Key, _Value> **p, bool& taller)
{
	switch ((*p)->BF)
	{
	case 1:
	{
		(*p)->BF = 0;
		taller = true;
		break;
	}
	case 0:
	{
		(*p)->BF = -1;
		taller = false;
		break;
	}
	case -1:
	{
		(*p)->BF = -2; // here need a rotate operation to rebalance since the node become unbalanced
		AVLNode<_Key, _Value> *q = (*p)->right;
		if (q->BF == 0) // operation is similer with RR rotate, but balance factors aren't the same as those after RR rotate
		{
			(*p)->right = q->left;
			q->left = *p;
			(*p)->BF = -1;
			q->BF = 1;
			*p = q;
			taller = false;
			/*
			 * E is the node to be erased, first step is erase, second step is rotate.
			 *        P                P                Q
			 *      /   \                \            /   \
			 *     E     Q    ->          Q    ->    P     R
			 *          / \              / \          \
			 *         L   R            L   R          L
			 * height of this subtree didn't change, thus taller is false.
			 */
		}
		else if (q->BF == -1)
		{
			*p = rotateRR(p);
			taller = true;
			/*
			 * E is the node to be erased, first step is erase, second step is rotate.
			 *        P                P                Q
			 *      /   \                \            /   \
			 *     E     Q    ->          Q    ->    P     R
			 *            \                \
			 *             R                R
			 * height of this subtree changed from 3 to 2, thus taller is true.
			 */
		}
		else // (q->BF == 1)
		{
			*p = rotateRL(p);
			taller = true;
			/*
			 * E is the node to be erased, first step is erase, second step is rotate.
			 *        P                P                L
			 *      /   \                \            /   \
			 *     E     Q    ->          Q    ->    P     Q
			 *          /                /
			 *         L                L
			 * height of this subtree changed from 3 to 2, thus taller is true.
			 */
		}
		break;
	}
	default:
		assert( false );
	}
}

template <typename _Key, typename _Value, typename _Compare>
void AVLTree<_Key, _Value, _Compare>::rightRebalance(AVLNode<_Key, _Value> **p, bool& taller)
{
	switch ((*p)->BF)
	{
	case -1:
	{
		(*p)->BF = 0;
		taller = true;
		break;
	}
	case 0:
	{
		(*p)->BF = 1;
		taller = false;
		break;
	}
	case 1:
	{
		(*p)->BF = 2; // here need a rotate operation to rebalance since the node become unbalanced
		AVLNode<_Key, _Value> *q = (*p)->left;
		if (q->BF == 0) // operation is similer with LL rotate, but balance factors aren't the same as those after LL rotate
		{
			(*p)->left = q->right;
			q->right = *p;
			(*p)->BF = 1;
			q->BF = -1;
			*p = q;
			taller = false;
			/*
			 * E is the node to be erased, first step is erase, second step is rotate.
			 *        P                P                Q
			 *      /   \            /                /   \
			 *     Q     E    ->    Q          ->    L     P
			 *    / \              / \                    /
			 *   L   R            L   R                  R
			 * height of this subtree didn't change, thus taller is false.
			 */
		}
		else if (q->BF == 1)
		{
			*p = rotateLL(p);
			taller = true;
			/*
			 * E is the node to be erased, first step is erase, second step is rotate.
			 *        P                P                Q
			 *      /   \            /                /   \
			 *     Q     E    ->    Q          ->    L     P
			 *    /                /
			 *   L                L
			 * height of this subtree changed from 3 to 2, thus taller is true.
			 */
		}
		else // (q->BF == -1)
		{
			*p = rotateLR(p);
			taller = true;
			/*
			 * E is the node to be erased, first step is erase, second step is rotate.
			 *        P                P                R
			 *      /   \            /                /   \
			 *     Q     E    ->    Q          ->    Q     P
			 *      \                \
			 *       R                R
			 * height of this subtree changed from 3 to 2, thus taller is true.
			 */
		}
		break;
	}
	default:
		assert( false );
	}
}

template <typename _Key, typename _Value, typename _Compare>
AVLNode<_Key, _Value>* AVLTree<_Key, _Value, _Compare>::rotateLL(AVLNode<_Key, _Value> **p)
{
	AVLNode<_Key, _Value> *q = (*p)->left;
	(*p)->left = q->right;
	q->right = *p;
	(*p)->BF = q->BF = 0;
	return q;
}

template <typename _Key, typename _Value, typename _Compare>
AVLNode<_Key, _Value>* AVLTree<_Key, _Value, _Compare>::rotateRR(AVLNode<_Key, _Value> **p)
{
	AVLNode<_Key, _Value> *q = (*p)->right;
	(*p)->right = q->left;
	q->left = *p;
	(*p)->BF = q->BF = 0;
	return q;
}

template <typename _Key, typename _Value, typename _Compare>
AVLNode<_Key, _Value>* AVLTree<_Key, _Value, _Compare>::rotateLR(AVLNode<_Key, _Value> **p)
{
	AVLNode<_Key, _Value> *q = (*p)->left;
	assert( (*p)->BF == 2 && q->BF == -1 );
	int flag = q->right->BF; // q->right->BF == { 1, -1, 0 }
	(*p)->left = rotateRR(&q);
	q->BF = flag == 1 ? 0 : (flag == -1 ? 1 : 0);
	AVLNode<_Key, _Value> *r = rotateLL(p); // r->BF == 0 always true
	(*p)->BF = flag == 1 ? -1 : (flag == -1 ? 0 : 0);
	return r;
}

template <typename _Key, typename _Value, typename _Compare>
AVLNode<_Key, _Value>* AVLTree<_Key, _Value, _Compare>::rotateRL(AVLNode<_Key, _Value> **p)
{
	AVLNode<_Key, _Value> *q = (*p)->right;
	assert( (*p)->BF == -2 && q->BF == 1 );
	int flag = q->left->BF; // q->left->BF == { 1, -1, 0 }
	(*p)->right = rotateLL(&q);
	q->BF = flag == 1 ? -1 : (flag == -1 ? 0 : 0);
	AVLNode<_Key, _Value> *r = rotateRR(p); // r->BF == 0 always true
	(*p)->BF = flag == 1 ? 0 : (flag == -1 ? 1 : 0);
	return r;
}

template <typename _Key, typename _Value, typename _Compare>
void AVLTree<_Key, _Value, _Compare>::preOrderTraverse(std::vector<AVLNode<_Key, _Value>* >& vec)
{
	if (node_count == 0) return;

	vec.reserve(node_count);

	std::stack<AVLNode<_Key, _Value>* > S;
	AVLNode<_Key, _Value> *p = root;
	while (p != NULL || !S.empty())
	{
		if (p != NULL)
		{
			vec.push_back(p);
			S.push(p);
			p = p->left;
		}
		else
		{
			p = S.top();
			S.pop();
			p = p->right;
		}
	}
}

template <typename _Key, typename _Value, typename _Compare>
void AVLTree<_Key, _Value, _Compare>::inOrderTraverse(std::vector<AVLNode<_Key, _Value>* >& vec)
{
	if (node_count == 0) return;

	vec.reserve(node_count);

	std::stack<AVLNode<_Key, _Value>* > S;
	AVLNode<_Key, _Value> *p = root;
	while (p != NULL || !S.empty())
	{
		if (p != NULL)
		{
			S.push(p);
			p = p->left;
		}
		else
		{
			p = S.top();
			S.pop();
			vec.push_back(p);
			p = p->right;
		}
	}
}

template <typename _Key, typename _Value, typename _Compare>
void AVLTree<_Key, _Value, _Compare>::postOrderTraverse(std::vector<AVLNode<_Key, _Value>* >& vec)
{
	if (node_count == 0) return;

	vec.reserve(node_count);

	std::stack<AVLNode<_Key, _Value>* > S;
	AVLNode<_Key, _Value> *prev = NULL, *cur = NULL;
	S.push(root);
	while (!S.empty())
	{
		cur = S.top();
		/* visit cur when it is a leaf node or its childs have been visited */
		if ( (cur->left == NULL && cur->right == NULL) || (prev != NULL && (prev == cur->left || prev == cur->right)) )
		{
			vec.push_back(cur);
			S.pop();
			prev = cur;
		}
		else
		{
			/* push() right child first, left child second, so left child comes first when pop() */
			if (cur->right != NULL)
				S.push(cur->right);
			if (cur->left != NULL)
				S.push(cur->left);
		}
	}
}

#endif /* __AVLTREE_H__ */

