#include "AVLTree.h"

#include <time.h>
#include <iostream>
#include <map>
#include <algorithm> // random_shuffle

using std::cout;
using std::endl;

/**
 * @brief convenience class, which prints procedure executing time between Timer's construction and its destruction.
 */
class Timer
{
public:
	/** record constructing time in the constructor. */
	explicit Timer(const char *s) : _note(s) { _start_time = clock(); }
	~Timer() { cout << _note << elapsed() << "s." << endl; }
	/** calculate how much time elapsed since constructed. */
	inline double elapsed() { return static_cast<double>(clock() - _start_time) / CLOCKS_PER_SEC; }

private:
	Timer(const Timer&);
	Timer& operator=(const Timer&);

private:
	std::string _note;   ///< this string will be printed when timer elapsed.
	clock_t _start_time; ///< record the time when the timer is constructed.
};

namespace nonstd
{

/**
 * @brief A non-standard map made up of (key,value) pairs, which can be retrieved based on a key, in logarithmic time.
 * 
 * @tparam _K  Type of key objects.
 * @tparam _V  Type of mapped objects.
 * @tparam _C  Comparison function object type, defaults to less<_K>.
 */
template <typename _K, typename _V, typename _C = std::less<_K> >
class map
{
public:
	/** @name container typedefs for easy of use. */
	///@{
	typedef _K key_type;
	typedef _V data_type;
	typedef _V mapped_type;
	typedef std::pair<const _K, _V> value_type;
	typedef _C key_compare;

	typedef AVLNode<_K, _V>* iterator;
	///@}

	/** @name constructors, destructor of nonstd::map. */
	///@{
	map() : _M_t() {}
	explicit map(const _C& __comp) : _M_t(__comp) {}
	map(const map& __x) : _M_t(__x._M_t) { }
	~map() { clear(); }
	///@}
	
	map& operator=(const map& __x) { _M_t = __x._M_t; return *this; }

	/** @name APIs of nonstd::map. */
	///@{
	bool empty() { return _M_t.empty(); }
	size_t size() { return _M_t.size(); }
	
	iterator begin() { return _M_t.minimum(); }
	iterator end() { return NULL; }

	iterator find(const key_type& __x) { return _M_t.find(__x); }
	size_t count(const key_type& __x) { return _M_t.find(__x) == end() ? 0 : 1; }

	iterator insert(const value_type& __x) { return _M_t.insert(__x.first, __x.second); }
	void erase(const key_type& __x) { _M_t.erase(__x); }

	void clear() { _M_t.clear(); }
	///@}

private:
	AVLTree<_K, _V, _C> _M_t; ///< AVL tree representing nonstd::map.
};

} // end namespace nonstd

#define DATA_SCALE    1000000
#define FAIL_RATIO    0.25

inline double dblrand() { return static_cast<double>(rand())/RAND_MAX + (static_cast<double>(rand())/RAND_MAX * (1.0/RAND_MAX)); }
inline int idxrand(int minIdx, int maxIdx) { return rand() % (maxIdx - minIdx) + minIdx; }

template <typename _Key, typename _Value>
void printTraverseResult(std::vector<AVLNode<_Key, _Value>* >& v)
{
	for (size_t i = 0; i < v.size(); ++i)
		cout << v[i]->key << ' ';
	cout << endl;
}

template <typename _Key, typename _Value>
void printBalanceFactor(std::vector<AVLNode<_Key, _Value>* >& v)
{
	for (size_t i = 0; i < v.size(); ++i)
		cout << v[i]->BF << ' ';
	cout << endl;
}

void fillAVLTree(AVLTree<int, double>& avlTree)
{
	avlTree.insert(1, 3.23);
	avlTree.insert(2, 1.34);
	avlTree.insert(3, 8.36);
	avlTree.insert(4, 9.73);
	avlTree.insert(5, 7.18);
	avlTree.insert(6, 7.51);
	avlTree.insert(7, 8.13);
	avlTree.insert(16, 4.16);
	avlTree.insert(15, 2.86);
	avlTree.insert(14, 1.01);
	avlTree.insert(13, 5.69);
	avlTree.insert(12, 4.97);
	avlTree.insert(11, 9.61);
	avlTree.insert(10, 3.27);
	avlTree.insert(8, 7.12);
	avlTree.insert(9, 8.69);
	/*
	 * After fill, the structure of AVL tree looks:
	 *                   7
	 * 
	 *              /         \
	 *
	 *            4             13
	 *         /     \       /      \
	 *        2       6     11      15
	 *       / \     /     /  \    /  \
	 *      1   3   5     9   12  14  16
	 *                   / \
	 *                  8  10
	 */
}

int main(int argc, char *argv[])
{
	cout << "Test AVLTree methods: " << endl;

	AVLTree<int, double> avlTree;
	fillAVLTree(avlTree);

	cout << "before erase :\n";
	std::vector<AVLNode<int, double>* > preOrderAVL, inOrderAVL, postOrderAVL;
	avlTree.preOrderTraverse(preOrderAVL);
	avlTree.inOrderTraverse(inOrderAVL);
	avlTree.postOrderTraverse(postOrderAVL);
	cout << "PreOrder: "; printTraverseResult(preOrderAVL);
	cout << "InOrder: "; printTraverseResult(inOrderAVL);
	cout << "BalanceFactor: "; printBalanceFactor(inOrderAVL);
	cout << "PostOrder: "; printTraverseResult(postOrderAVL);
	preOrderAVL.clear(); inOrderAVL.clear(); postOrderAVL.clear();
	
	avlTree.erase(13);
	
	/*
	 * After erase 13, the structure of AVL tree looks:
	 *                   7
	 * 
	 *              /         \
	 *
	 *            4             12
	 *         /     \       /      \
	 *        2       6     9       15
	 *       / \     /     /  \    /  \
	 *      1   3   5     8   11  14  16
	 *                       /
	 *                     10
	 */
	cout << "after erase :\n";
	avlTree.preOrderTraverse(preOrderAVL);
	avlTree.inOrderTraverse(inOrderAVL);
	avlTree.postOrderTraverse(postOrderAVL);
	cout << "PreOrder: "; printTraverseResult(preOrderAVL);
	cout << "InOrder: "; printTraverseResult(inOrderAVL);
	cout << "BalanceFactor: "; printBalanceFactor(inOrderAVL);
	cout << "PostOrder: "; printTraverseResult(postOrderAVL);
	
	avlTree.clear();

	/*
	cout << "performance evaluation of AVLTree, compared with std::map which implemented by red-black tree.\n";
	std::map<double, double> std_map;
	nonstd::map<double, double> nonstd_map;

	// unsigned int seed = 1;
	srand(time(NULL));
	// srand(seed);
	
	int i = 0;
	// generate random keys and values
	std::vector<int> keys;
	std::vector<double> values;
	keys.reserve(DATA_SCALE);
	values.reserve(DATA_SCALE);
	for (i = 0; i < DATA_SCALE; ++i)
	{
		keys.push_back(rand());
		values.push_back(dblrand());
	}
	
	// determine the order of elements' index in the operations of insert(), erase() and find()
	std::vector<int> insertOrder, eraseOrder, findOrder;
	insertOrder.reserve(2*DATA_SCALE);
	eraseOrder.reserve(DATA_SCALE);
	findOrder.reserve(DATA_SCALE);
	for (i = 0; i < DATA_SCALE; ++i) // 0-100%
	{
		insertOrder.push_back(i); // 100%
		if (i % 4 == 0)
			eraseOrder.push_back(i); // 25%
		else if (i % 4 == 1)
			findOrder.push_back(i); // 25%
	}
	for (i = 0; i < FAIL_RATIO*DATA_SCALE; ++i)
	{
		int idx = idxrand(0, DATA_SCALE);
		insertOrder.push_back(idx); // FAIL_RATIO * 100%
	}
	for (i = 0; i < FAIL_RATIO*DATA_SCALE; i += 2)
	{
		int idx = idxrand(0, DATA_SCALE);
		while (idx % 4 < 2)
			idx = idxrand(0, DATA_SCALE);
		eraseOrder.push_back(idx); // FAIL_RATIO * 50%
	}
	for (i = 0; i < FAIL_RATIO*DATA_SCALE; i += 2)
	{
		int idx = idxrand(0, DATA_SCALE);
		while (idx % 4 < 2)
			idx = idxrand(0, DATA_SCALE);
		findOrder.push_back(idx); // FAIL_RATIO * 50%
	}

	std::random_shuffle(insertOrder.begin(), insertOrder.end()); // using built-in random generator
	std::random_shuffle(eraseOrder.begin(), eraseOrder.end()); // using built-in random generator
	std::random_shuffle(findOrder.begin(), findOrder.end()); // using built-in random generator

	// execute operations of insert(), erase() and find() for both std::map and nonstd::map
	{
		Timer timer("insertion of std::map costs: ");
		for (i = 0; i < DATA_SCALE * (1+FAIL_RATIO); ++i)
			std_map.insert(std::pair<int, double>(keys[insertOrder[i]], values[insertOrder[i]]));
	}
	{
		Timer timer("insertion of nonstd::map costs: ");
		for (i = 0; i < DATA_SCALE * (1+FAIL_RATIO); ++i)
			nonstd_map.insert(std::pair<int, double>(keys[insertOrder[i]], values[insertOrder[i]]));
	}
	{
		Timer timer("erasure of std::map costs: ");
		for (i = 0; i < DATA_SCALE/2 * (1+FAIL_RATIO); ++i)
			std_map.erase(keys[eraseOrder[i]]);
	}
	{
		Timer timer("erasure of nonstd::map costs: ");
		for (i = 0; i < DATA_SCALE/2 * (1+FAIL_RATIO); ++i)
			nonstd_map.erase(keys[eraseOrder[i]]);
	}
	{
		Timer timer("finding of std::map costs: ");
		int existCount = 0;
		for (i = 0; i < DATA_SCALE/2 * (1+FAIL_RATIO); ++i)
			if (std_map.find(keys[findOrder[i]]) != std_map.end())
				++existCount;
		cout << "std::map::find() exist elements' count: " << existCount << endl;
	}
	{
		Timer timer("finding of nonstd::map costs: ");
		int existCount = 0;
		for (i = 0; i < DATA_SCALE/2 * (1+FAIL_RATIO); ++i)
			if (nonstd_map.find(keys[findOrder[i]]) != nonstd_map.end())
				++existCount;
		cout << "nonstd::map::find() exist elements' count: " << existCount << endl;
	}

	std_map.clear();
	nonstd_map.clear();

	cout << "From comparsion, we can draw a conclusion that the performance of nonstd::map is slightly better than std::map." << endl;
	*/
	return 0;
}

